//camera implementation
//Version 2.0

#include "camera.h"
#include "input.h"
#include "game.h"


//Camera class constructor implementation
Camera::Camera()
{
	//Reset the identity matrix for each of the camera matrices
	D3DXMatrixIdentity(&mView);
	D3DXMatrixIdentity(&mProj);
	D3DXMatrixIdentity(&mVP);

	//Assign default vector values
	mPos	=	D3DXVECTOR3(0.0f, 0.0f, -5.0f);
	mRight	=	D3DXVECTOR3(1.0f, 0.0f, 0.0f);
	mUp		=	D3DXVECTOR3(0.0f, 1.0f, 0.0f);
	mLook	=	D3DXVECTOR3(0.0f, 0.0f, 1.0f);

	//Set default camera speed
	mSpeed = 10.0f;
}

Camera::~Camera()
{

}

void Camera::SetPerspective(float fov, float aspect, float nearZ, float farZ)
{
	D3DXMatrixPerspectiveFovLH(&mProj, fov, aspect, nearZ, farZ);
	mVP = mView * mProj;
}

void Camera::SetSpeed(float speed)
{
	mSpeed = speed;
}

void Camera::BuildView()
{
	//Keep camera's axes orthogonal to each other and of unit length.
	D3DXVec3Normalize(&mLook, &mLook);

	D3DXVec3Cross(&mUp, &mLook, &mRight);
	D3DXVec3Normalize(&mUp, &mUp);

	D3DXVec3Cross(&mRight, &mUp, &mLook);
	D3DXVec3Normalize(&mRight, &mRight);

	//Fill in the view matrix entries.
	float x = -D3DXVec3Dot(&mPos, &mRight);
	float y = -D3DXVec3Dot(&mPos, &mUp);
	float z = -D3DXVec3Dot(&mPos, &mLook);

	mView(0,0) = mRight.x; 
	mView(1,0) = mRight.y; 
	mView(2,0) = mRight.z; 
	mView(3,0) = x;   

	mView(0,1) = mUp.x;
	mView(1,1) = mUp.y;
	mView(2,1) = mUp.z;
	mView(3,1) = y;  

	mView(0,2) = mLook.x; 
	mView(1,2) = mLook.y; 
	mView(2,2) = mLook.z; 
	mView(3,2) = z;   

	mView(0,3) = 0.0f;
	mView(1,3) = 0.0f;
	mView(2,3) = 0.0f;
	mView(3,3) = 1.0f;
}

void Camera::Update(float dTime)
{
	//Default direction vector
	D3DXVECTOR3 dir(0.0f, 0.0f, 0.0f);

	//Get camera input here
	if(gGameInput->GetKeyPress(DIK_W, true)) //forward
	{
		dir += mLook;
	}
	else if (gGameInput->GetKeyPress(DIK_S, true)) //backward
	{
		dir -= mLook;
	}
	else if (gGameInput->GetKeyPress(DIK_D, true)) //right
	{
		dir += mRight;
	}
	else if (gGameInput->GetKeyPress(DIK_A, true)) //left
	{
		dir -= mRight;
	}

	//Move camera along normalized direction
	D3DXVec3Normalize(&dir, &dir);
	mPos += dir * mSpeed * dTime;

	// We rotate at a fixed speed.
	float pitch  = gGameInput->GetMouseDeltaY() / 300.0f;
	float yAngle = gGameInput->GetMouseDeltaX() / 300.0f;

	// Rotate camera's look and up vectors around the camera's right vector.
	D3DXMATRIX R;
	D3DXMatrixRotationAxis(&R, &mRight, pitch);
	D3DXVec3TransformCoord(&mLook, &mLook, &R);
	D3DXVec3TransformCoord(&mUp, &mUp, &R);

	// Rotate camera axes about the world's y-axis.
	D3DXMatrixRotationY(&R, yAngle);
	D3DXVec3TransformCoord(&mRight, &mRight, &R);
	D3DXVec3TransformCoord(&mUp, &mUp, &R);
	D3DXVec3TransformCoord(&mLook, &mLook, &R);

	//Rebuild the view matrix every frame
	BuildView();
	mVP = mView * mProj;
}
