//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Direct 3D Framework
// Written by Vladeta Stojanovic (0602920@live.abertay.ac.uk)
// Based on example code from Introduction to 3D Game Programming with DirectX 9.0c: A Shader Based Approach by Frank D. Luna
// and the code provided by Dr Ruth Falconer.
//
// Version: 061210_0.9
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "game.h"
#include "vertex.h"
#include "camera.h"

//Global game framework instance variables
Timer* gTimer = NULL;
Camera* gCamera = NULL;
Scene* gScene = NULL;

//Constructor implementation
TestApp::TestApp(HINSTANCE hInstance, std::string winCaption, D3DDEVTYPE devType, DWORD requestedVP): 
D3DWindow(hInstance, winCaption, devType, requestedVP)
{
	//seed random time generator (used for changing to random font colors)
	srand(time_t(0));

	//Initialize the input & sound systems
	gGameInput = new Input(getWinHwnd());
	gTimer = new Timer();
	gCamera = new Camera();
	gScene = new Scene();
	
	//Initialize default scene camera values here
	mCamPos.x = 10.0f;
	mCamPos.y = 5.0f;
	mCamPos.z = -10.f;

	//if device failed...
	if(!checkDeviceCaps())
	{
		MessageBox(0, "checkDeviceCaps() Failed. Most likely no VS & PS v2.0 support...", 0, 0);
		PostQuitMessage(0);
	}

	//set up the font
	D3DXFONT_DESC fontDesc;
	fontDesc.Height          = 80;
    fontDesc.Width           = 40;
    fontDesc.Weight          = 0;
    fontDesc.MipLevels       = 0;
    fontDesc.Italic          = false;
    fontDesc.CharSet         = DEFAULT_CHARSET;
    fontDesc.OutputPrecision = OUT_DEFAULT_PRECIS;
    fontDesc.Quality         = DEFAULT_QUALITY;
    fontDesc.PitchAndFamily  = DEFAULT_PITCH | FF_DONTCARE;
    strcpy_s(fontDesc.FaceName, _T("OCR-B 10 BT"));

	//send info to debugger
	HR(D3DXCreateFontIndirect(gd3dDevice, &fontDesc, &mFont));

	onResetDevice();
}

//Destructor implementation
TestApp::~TestApp()
{
	Release(mFont);
	delete gTimer;
}


//Check device caps implementation (target graphics card must support vs & ps version 2.0)
bool TestApp::checkDeviceCaps()
{
	D3DCAPS9 caps;
	HR(gd3dDevice->GetDeviceCaps(&caps));

	if(caps.VertexShaderVersion < D3DVS_VERSION(2, 0))
	{
		return false;
	}

	if(caps.PixelShaderVersion < D3DPS_VERSION(2, 0))
	{
		return false;
	}

	return true;
}

//Overloaded on lost device function
void TestApp::onLostDevice()
{
	HR(mFont->OnLostDevice());
	gTimer->TimerOnLostDevice();
	gScene->SceneOnLostDevice();
}

//Overloaded on reset device function
void TestApp::onResetDevice()
{
	HR(mFont->OnResetDevice());
	gTimer->TimerOnResetDevice();
	gScene->SceneOnResetDevice();

	BuildProjMatrix();
}

//Overloaded update scene function
void TestApp::Update(float fTime)
{
	//Update input
	gGameInput->UpdateInput();

	//Timer specific
	gTimer->TimerUpdate(fTime);

	//Update the camera
	gCamera->Update(fTime);

	//Update water
	gScene->UpdateScene(fTime);

	//build the view matrix
	BuildViewMatrix();
}

//Overloaded draw scene function
void TestApp::Draw() 
{
	// Set source blend flag to use source alpha 
	HR(gd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA));

	// Set destination blend flag to use inverse source alpha
	HR(gd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA));

	HR(gd3dDevice->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(148, 148, 148), 1.0f, 0));
	HR(gd3dDevice->BeginScene());

		gScene->Draw(gCamera->GetVP(), gCamera->GetLook());
		gTimer->TimerDisplay();

	HR(gd3dDevice->EndScene());
	HR(gd3dDevice->Present(0, 0, 0, 0));
}

//Build view matrix
void TestApp::BuildViewMatrix()
{
	D3DXVECTOR3 defTarget(10.0f, 10.0f, 0.0f); //default view target
	D3DXVECTOR3 defUp(0.0f, 1.0f, 0.0f); //default up
	D3DXMatrixLookAtLH(&mView, &mCamPos, &defTarget, &defUp); //add updated camera coords here
}

//Build projection matrix
void TestApp::BuildProjMatrix()
{
	float w = (float)md3dPP.BackBufferWidth;
	float h = (float)md3dPP.BackBufferHeight;
	gCamera->SetPerspective(D3DX_PI * 0.25f, w/h, 1.0f, 5000.0f);
}

/////////////
//WinMain()
/////////////
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE prevInstance, PSTR cmdLine, int showCmd)
{
	// Enable run-time memory check for debug builds.
	#if defined(DEBUG) | defined(_DEBUG)
		_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	#endif

	TestApp Test(hInstance, "D3D Framework Test Application by Vladeta Stojanovic", D3DDEVTYPE_HAL, D3DCREATE_HARDWARE_VERTEXPROCESSING);
	gd3dWin = &Test;

	return gd3dWin->Run();
}