///////////////////////////////
// Mesh class implementation
///////////////////////////////

//todo: remove texture loading/extraction function from the loadxfile() routine.

#include "mesh.h"
#include "vertex.h"

using namespace std;

XMesh::XMesh(void)
{
	
}

XMesh::~XMesh()
{
	Release(mMesh);
	Release(mLevel);

	for(int i = 0; i < mTex.size(); ++i)
	{
		Release(mTex[i]);
	}

	Release(mDefTex);
	Release(mEnvMap);
	Release(mNMTex);
}

void XMesh::LoadXFile(const std::string& filename, 
					ID3DXMesh** meshOut,
					std::vector<Mtrl>& mtrls, 
					std::vector<IDirect3DTexture9*>& texs)
{
	// Step 1: Load the .x file from file into a system memory mesh.

	ID3DXMesh* meshSys      = 0;
	ID3DXBuffer* adjBuffer  = 0;
	ID3DXBuffer* mtrlBuffer = 0;
	DWORD numMtrls          = 0;

	HR(D3DXLoadMeshFromX(filename.c_str(), D3DXMESH_SYSTEMMEM, gd3dDevice,
		&adjBuffer,	&mtrlBuffer, 0, &numMtrls, &meshSys));


	// Step 2: Find out if the mesh already has normal info?
	D3DVERTEXELEMENT9 elems[MAX_FVF_DECL_SIZE];
	HR(meshSys->GetDeclaration(elems));
	
	bool hasNormals = false;
	D3DVERTEXELEMENT9 term = D3DDECL_END();
	for(int i = 0; i < MAX_FVF_DECL_SIZE; ++i)
	{
		// Did we reach D3DDECL_END() {0xFF,0,D3DDECLTYPE_UNUSED, 0,0,0}?
		if(elems[i].Stream == 0xff )
			break;

		if( elems[i].Type == D3DDECLTYPE_FLOAT3 &&
			elems[i].Usage == D3DDECLUSAGE_NORMAL &&
			elems[i].UsageIndex == 0 )
		{
			hasNormals = true;
			break;
		}
	}


	// Step 3: Change vertex format to VertexPNT.

	D3DVERTEXELEMENT9 elements[64];
	UINT numElements = 0;
	VertexPNT::Decl->GetDeclaration(elements, &numElements);

	ID3DXMesh* temp = 0;
	HR(meshSys->CloneMesh(D3DXMESH_SYSTEMMEM, 
		elements, gd3dDevice, &temp));
	Release(meshSys);
	meshSys = temp;


	// Step 4: If the mesh did not have normals, generate them.

	if( hasNormals == false)
		HR(D3DXComputeNormals(meshSys, 0));


	// Step 5: Optimize the mesh.

	HR(meshSys->Optimize(D3DXMESH_MANAGED | 
	D3DXMESHOPT_COMPACT | D3DXMESHOPT_ATTRSORT | D3DXMESHOPT_VERTEXCACHE, 
	(DWORD*)adjBuffer->GetBufferPointer(), 0, 0, 0, meshOut));

	D3DXCleanMesh(D3DXCLEAN_BOWTIES, meshSys, (DWORD*)adjBuffer, meshOut, 0, 0); //Why isn't this working?!

	Release(meshSys); // Done w/ system mesh.
	Release(adjBuffer); // Done with buffer.

	// Step 6: Extract the materials and load the textures.

	if( mtrlBuffer != 0 && numMtrls != 0 )
	{
		D3DXMATERIAL* d3dxmtrls = (D3DXMATERIAL*)mtrlBuffer->GetBufferPointer();

		for(DWORD i = 0; i < numMtrls; ++i)
		{
			// Save the ith material.  Note that the MatD3D property does not have an ambient
			// value set when its loaded, so just set it to the diffuse value.
			Mtrl m;
			m.ambient   = d3dxmtrls[i].MatD3D.Diffuse;
			m.diffuse   = d3dxmtrls[i].MatD3D.Diffuse;
			m.spec      = d3dxmtrls[i].MatD3D.Specular;
			m.specPower = d3dxmtrls[i].MatD3D.Power;
			mtrls.push_back( m );

			// Check if the ith material has an associative texture
			if( d3dxmtrls[i].pTextureFilename != 0 )
			{
				// Yes, load the texture for the ith subset
				IDirect3DTexture9* tex = 0;
				char* texFN = d3dxmtrls[i].pTextureFilename;
				D3DXCreateTextureFromFile(gd3dDevice, texFN, &tex);

				// Save the loaded texture
				texs.push_back( tex );
			}
			else
			{
				// No texture for the ith subset
				texs.push_back( 0 );
			}
		}
	}
	Release(mtrlBuffer); // done w/ buffer
}

void XMesh::GenTexCoords(float texScale)
{
	if(texScale != NULL)
	{
		D3DVERTEXELEMENT9 elements[64];
		UINT numElements = 0;

		VertexPNT::Decl->GetDeclaration(elements, &numElements);

		ID3DXMesh* temp = 0;
		HR(mMesh->CloneMesh(D3DXMESH_SYSTEMMEM, elements, gd3dDevice, &temp));

		Release(mMesh);

		// Now generate texture coordinates for each vertex.
		VertexPNT* vertices = 0;
		HR(temp->LockVertexBuffer(0, (void**)&vertices));

		for(UINT i = 0; i < temp->GetNumVertices(); ++i)
		{
			// Convert to spherical coordinates.
			D3DXVECTOR3 p = vertices[i].pos;
		
			float theta = atan2f(p.z, p.x);
			float phi   = acosf(p.y / sqrtf(p.x*p.x+p.y*p.y+p.z*p.z));

			// Phi and theta give the texture coordinates, but are not in 
			// the range [0, 1], so scale them into that range.
			float u = theta / (texScale * D3DX_PI);
			float v = phi   / D3DX_PI;
			
			// Save texture coordinates.
			
			vertices[i].tex0.x = u;
			vertices[i].tex0.y = v;
		}
		HR(temp->UnlockVertexBuffer());

		// Clone back to a hardware mesh.
		HR(temp->CloneMesh(D3DXMESH_MANAGED | D3DXMESH_WRITEONLY, elements, gd3dDevice, &mMesh));

		Release(temp);
	}
}

//load and init mesh with diffuse + environment map
void XMesh::LoadandInitEnvMap(char * meshFile, char * meshTex, char * envTex, float texScale_)
{
	LoadXFile(meshFile, &mMesh, mMtrl, mTex);
	D3DXCreateTextureFromFile(gd3dDevice, meshTex, &mDefTex);
	D3DXCreateCubeTextureFromFile(gd3dDevice,  envTex, &mEnvMap);
	GenTexCoords(texScale_);
}

//load and init mesh with diffuse texture only
void XMesh::LoadandInitDiffMap(char * meshFile, char * meshTex,float texScale_)
{
	//implement me when needed
	GenTexCoords(texScale_);
}

//load and init mesh with diffuse + normal map
void XMesh::LoadandInitNMMap(char * meshFile, char * meshTex, char * nmTex)
{
	ID3DXMesh* tempMesh = 0;

	LoadXFile(meshFile, &tempMesh, mMtrl, mTex);
	D3DXCreateTextureFromFile(gd3dDevice, nmTex, &mNMTex);
	D3DXCreateTextureFromFile(gd3dDevice, meshTex, &mLevelTex);

	D3DVERTEXELEMENT9 elems[MAX_FVF_DECL_SIZE];
	UINT numElems = 0;
	HR(NMapVertex::Decl->GetDeclaration(elems, &numElems));

	// Clone the mesh to the NMapVertex format
	ID3DXMesh* clonedTempMesh = 0;
	HR(tempMesh->CloneMesh(D3DXMESH_MANAGED, elems, gd3dDevice, &clonedTempMesh));

	// Now use D3DXComputeTangentFrameEx to build the TNB-basis for each vertex
	// in the mesh
	HR(D3DXComputeTangentFrameEx(
	  clonedTempMesh, 
	  D3DDECLUSAGE_TEXCOORD, 0, 
      D3DDECLUSAGE_BINORMAL, 0, 
	  D3DDECLUSAGE_TANGENT, 0,  
      D3DDECLUSAGE_NORMAL, 0,   
      0, 
      0, 
	  0.01f, 0.25f, 0.01f, 
	  &mLevel, 
	  0));        
	// Done with temps.
	Release(tempMesh);
	Release(clonedTempMesh);

	//GenTexCoords(texScale_);
}


void XMesh::DrawEnvBlnd(D3DXMATRIX* pMatrix, ID3DXEffect* pFX, D3DXHANDLE phMtrl, D3DXHANDLE phTex0, D3DXHANDLE phTex1)
{
	UINT numPasses = 0;
	HR(pFX->Begin(&numPasses, 0));
	HR(pFX->BeginPass(0));

	for(int j = 0; j < mMtrl.size(); ++j)
	{
		HR(pFX->SetValue(phMtrl, &mMtrl[j], sizeof(Mtrl)));
	
		// If there is a texture, then use.
		if(mTex[j] != 0)
		{
			HR(pFX->SetTexture(phTex1, mTex[j]));
		}
		// If no texture, use default white texture
		else
		{
			//Todo: add multitexturing support for blending up to two different texture layers.
			HR(pFX->SetTexture(phTex0, mEnvMap));
			HR(pFX->SetTexture(phTex1, mDefTex));
		}
	
		HR(pFX->CommitChanges());
		HR(mMesh->DrawSubset(j));
	}

	HR(pFX->EndPass());
	HR(pFX->End());
}

void XMesh::DrawDiffNM(D3DXMATRIX *pMatrix, ID3DXEffect *pFX, D3DXHANDLE phNormalMap, D3DXHANDLE phMtrl, D3DXHANDLE phTex)
{
	UINT numPasses = 0;
	HR(pFX->Begin(&numPasses, 0));
	HR(pFX->BeginPass(0));

	for(int j = 0; j < mMtrl.size(); ++j)
	{
		HR(pFX->SetValue(phMtrl, &mMtrl[j], sizeof(Mtrl)));
		HR(pFX->SetTexture(phTex, mLevelTex));
		HR(pFX->SetTexture(phNormalMap,  mNMTex));

		HR(pFX->CommitChanges());
		HR(mLevel->DrawSubset(j));
	}

	HR(pFX->EndPass());
	HR(pFX->End());
}

void XMesh::AnimRotX(float fTime)
{
	float xRot = 0.3f * fTime;

	VertexPNT* vertices = 0;

	HR(mMesh->LockVertexBuffer(0, (void**)&vertices));

	for(UINT i = 0; i < mMesh->GetNumVertices(); ++i)
	{
		D3DXMATRIX R;
		D3DXMatrixRotationYawPitchRoll(&R, xRot, 0.0f, 0.0f);
		D3DXVec3TransformCoord(&vertices[i].pos, &vertices[i].pos, &R);
	}
	HR(mMesh->UnlockVertexBuffer());
}