//Scene implementation

#include "scene.h"
#include "vertex.h"
#include "mesh.h"
#include "sky.h"
#include "water.h"

XMesh* gMesh = NULL;
XMesh* gLevel = NULL;
Sky* gSky = NULL;
Water* gWater = NULL;
float reflectivity1 = 0.9f;

Scene::Scene()
{
	InitVertDeclerations();

	//Material/Lighting properties
	mLight.dirW    = D3DXVECTOR3(1.0f, 100.0f, -1.0f);
	D3DXVec3Normalize(&mLight.dirW, &mLight.dirW);
	mLight.ambient = D3DXCOLOR(1.0f, 1.0f, 1.0f, 1.0f);
	mLight.diffuse = D3DXCOLOR(1.0f, 1.0f, 1.0f, 1.0f);
	mLight.spec    = D3DXCOLOR(0.8f, 0.8f, 0.8f, 1.0f);

	//Setup the skysphere and the mesh class instances
	gMesh = new XMesh();
	gLevel = new XMesh();
	gSky = new Sky("models/scene/envmap2.dds", 10000.0f);
	

	////////////////////////////////
	//Water plane rendering setup
	////////////////////////////////
	D3DXMATRIX waterWorld;
	D3DXMatrixTranslation(&waterWorld, 0.0f, -3.0f, 0.0f);

	Mtrl waterMtrl;
	waterMtrl.ambient   = D3DXCOLOR(0.26f, 0.23f, 0.3f, 0.90f);
	waterMtrl.diffuse   = D3DXCOLOR(0.26f, 0.23f, 0.3f, 0.90f);
	waterMtrl.spec      = 255;
	waterMtrl.specPower = 64.0f;

	WaterInitInfo waterInitInfo;
	waterInitInfo.dirLight = mLight;
	waterInitInfo.mtrl     = waterMtrl;
	waterInitInfo.vertRows         = 128;
	waterInitInfo.vertCols         = 128;
	waterInitInfo.dx               = 0.25f;
	waterInitInfo.dz               = 0.19f;
	waterInitInfo.waveMapFilename0 = "models/scene/wave0.dds";
	waterInitInfo.waveMapFilename1 = "models/scene/wave1.dds";
	waterInitInfo.waveMapVelocity0 = D3DXVECTOR2(0.05f, 0.08f);
	waterInitInfo.waveMapVelocity1 = D3DXVECTOR2(-0.02f, 0.1f);
	waterInitInfo.texScale = 8.0f;
	waterInitInfo.toWorld = waterWorld;
 
	gWater = new Water(waterInitInfo);
	gWater->setEnvMap(gSky->getEnvMap());

	//Load .x file and the associated texture
	gMesh->LoadandInitEnvMap("models/scene/skullocc.x", "models/scene/skull_tex.tga", "models/scene/envmap2.dds" ,  0.5f);
	gLevel->LoadandInitNMMap("models/scene/pool.x", "models/scene/pool_tex.tga" , "models/scene/pool_tex_nm.tga"); 

	D3DXMatrixIdentity(&mWorld);
	D3DXMatrixIdentity(&mWorldInv);

	BuildFX();

	SceneOnResetDevice();
}

Scene::~Scene()
{
	Release(mFX0);
	Release(mFX1);

	DestroyVertDeclerations();
}

void Scene::SceneOnLostDevice()
{
	HR(mFX0->OnLostDevice());
	HR(mFX1->OnLostDevice());
	gSky->OnLostDevice();
	gWater->OnLostDevice();
}

void Scene::SceneOnResetDevice()
{
	HR(mFX0->OnResetDevice());
	HR(mFX1->OnResetDevice());
	gSky->OnLostDevice();
	gWater->OnResetDevice();
}

void Scene::UpdateScene(float dTime)
{
	gWater->update(dTime);
	gMesh->AnimRotX(dTime);
}

void Scene::Draw(D3DXMATRIX pMatrix, D3DXVECTOR3 CamPos)
{
	// Setup the rendering FX
	HR(mFX0->SetTechnique(mhFXTech0));

	mFX0->SetValue(mhLight, &mLight, sizeof(DirLight));
	mFX0->SetValue(mhReflectivity, &reflectivity1, sizeof(float));

	D3DXMATRIX worldInvTrans;
	D3DXMatrixInverse(&worldInvTrans, 0, &pMatrix);
	D3DXMatrixTranspose(&worldInvTrans, &worldInvTrans);

	mFX0->SetMatrix(mhWVP, &pMatrix);
	mFX0->SetMatrix(mhEyePos, &pMatrix);
	mFX0->SetMatrix(mhWorldInvTrans, &pMatrix);
	mFX0->SetMatrix(mhWorld, &pMatrix);
	mFX0->SetValue(mhWorld, &pMatrix, sizeof(D3DXVECTOR3));

	//Draw skysphere here
	gSky->draw(pMatrix, CamPos);

	//Draw the water here
	gWater->draw(pMatrix, CamPos);

	//Draw environment mapped skull mesh here
	gMesh->DrawEnvBlnd(&pMatrix, mFX0, mhMtrl,  mhEnvMap, mhTex);

	//Draw normal mapped scene geometry here
	mFX1->SetValue(mhLightfx1, &mLight, sizeof(DirLight));
	mFX1->SetMatrix(mhWorldInv, &mWorldInv);
	mFX1->SetMatrix(mhEyePosWfx1, &pMatrix);
	mFX1->SetMatrix(mhWVPfx1, &pMatrix);
	mFX1->SetTechnique(mhFXTech1);

	gLevel->DrawDiffNM(&pMatrix, mFX1, mhNormalMap, mhMtrlfx1, mhTexfx1);
}

void Scene::BuildFX()
{
	// Create the FX from a .fx file.
	ID3DXBuffer* errors = 0;
	HR(D3DXCreateEffectFromFile(gd3dDevice, "shaders/EnvMap.c", //if changing shader, change shader dir here
		0, 0, D3DXSHADER_DEBUG, 0, &mFX0, &errors));

	HR(D3DXCreateEffectFromFile(gd3dDevice, "shaders/NormalMap.c",
		0, 0, D3DXSHADER_DEBUG, 0, &mFX1, &errors));

	if(errors)
	{
		MessageBox(0, (char*)errors->GetBufferPointer(), 0, 0);
	}

	//Obtain handles ofr environment map shader
	mhFXTech0               = mFX0->GetTechniqueByName("EnvMapTech"); //and always check that you are calling the correct technique
	mhWVP                   = mFX0->GetParameterByName(0, "gWVP");
	mhWorldInvTrans			= mFX0->GetParameterByName(0, "gWorldInvTrans");
	mhMtrl					= mFX0->GetParameterByName(0, "gMtrl");
	mhLight					= mFX0->GetParameterByName(0, "gLight");
	mhEyePos				= mFX0->GetParameterByName(0, "gEyePosW");
	mhWorld					= mFX0->GetParameterByName(0, "gWorld");
	mhTex					= mFX0->GetParameterByName(0, "gTex");
	mhEnvMap				= mFX0->GetParameterByName(0, "gEnvMap");
	mhReflectivity			= mFX0->GetParameterByName(0, "gReflectivity");


	//Obtain handles for normal map shader
	mhFXTech1    = mFX1->GetTechniqueByName("NormalMapTech");
	mhWVPfx1        = mFX1->GetParameterByName(0, "gWVP");
	mhWorldInv   = mFX1->GetParameterByName(0, "gWorldInv");
	mhMtrlfx1       = mFX1->GetParameterByName(0, "gMtrl");
	mhLightfx1      = mFX1->GetParameterByName(0, "gLight");
	mhEyePosWfx1    = mFX1->GetParameterByName(0, "gEyePosW");
	mhTexfx1        = mFX1->GetParameterByName(0, "gTex");
	mhNormalMap  = mFX1->GetParameterByName(0, "gNormalMap");
}
