//Water plane rendering interface implementation

#include "water.h"
#include "vertex.h"

Water::Water(WaterInitInfo& initInfo)
{
	mInitInfo = initInfo;

	mWidth = (initInfo.vertCols-1)*initInfo.dx;
	mDepth = (initInfo.vertRows-1)*initInfo.dz;

	mWaveMapOffset0 = D3DXVECTOR2(0.0f, 0.0f);
	mWaveMapOffset1 = D3DXVECTOR2(0.0f, 0.0f);

	DWORD numTris  = (initInfo.vertRows-1)*(initInfo.vertCols-1)*2;
	DWORD numVerts = initInfo.vertRows*initInfo.vertCols;


	//===============================================================
	// Allocate the mesh.


	D3DVERTEXELEMENT9 elems[MAX_FVF_DECL_SIZE];
	UINT numElems = 0;
	HR(VertexPNT::Decl->GetDeclaration(elems, &numElems));
	HR(D3DXCreateMesh(numTris, numVerts, 
		D3DXMESH_MANAGED, elems, gd3dDevice, &mMesh));


	//===============================================================
	// Write the grid vertices and triangles to the mesh.

	VertexPNT* v = 0;
	HR(mMesh->LockVertexBuffer(0, (void**)&v));
	
	std::vector<D3DXVECTOR3> verts;
	std::vector<DWORD> indices;
	GenerateGrid(mInitInfo.vertRows, mInitInfo.vertCols, mInitInfo.dx, 
		mInitInfo.dz, D3DXVECTOR3(0.0f, 0.0f, 0.0f), verts, indices);

	for(int i = 0; i < mInitInfo.vertRows; ++i)
	{
		for(int j = 0; j < mInitInfo.vertCols; ++j)
		{
			DWORD index   = i * mInitInfo.vertCols + j;
			v[index].pos  = verts[index];
			v[index].tex0 = D3DXVECTOR2((float)j/mInitInfo.vertCols, 
				                        (float)i/mInitInfo.vertRows)
										* initInfo.texScale;
		}
	}
	HR(mMesh->UnlockVertexBuffer());

	//===============================================================
	// Write triangle data so we can compute normals.

	WORD* indexBuffPtr = 0;
	HR(mMesh->LockIndexBuffer(0, (void**)&indexBuffPtr));
	DWORD* attBuff = 0;
	HR(mMesh->LockAttributeBuffer(0, &attBuff));
	for(UINT i = 0; i < mMesh->GetNumFaces(); ++i)
	{
		indexBuffPtr[i*3+0] = (WORD)indices[i*3+0];
		indexBuffPtr[i*3+1] = (WORD)indices[i*3+1];
		indexBuffPtr[i*3+2] = (WORD)indices[i*3+2];

		attBuff[i] = 0; // All in subset 0.
	}
	HR(mMesh->UnlockIndexBuffer());
	HR(mMesh->UnlockAttributeBuffer());

	//===============================================================
	// Optimize for the vertex cache and build attribute table.

	DWORD* adj = new DWORD[mMesh->GetNumFaces()*3];
	HR(mMesh->GenerateAdjacency(EPSILON, adj));
	HR(mMesh->OptimizeInplace(D3DXMESHOPT_VERTEXCACHE|D3DXMESHOPT_ATTRSORT,
		adj, 0, 0, 0));
	delete[] adj;


	//===============================================================
	// Create textures/effect.

	HR(D3DXCreateTextureFromFile(gd3dDevice, initInfo.waveMapFilename0, &mWaveMap0));
	HR(D3DXCreateTextureFromFile(gd3dDevice, initInfo.waveMapFilename1, &mWaveMap1));
	buildFX();
}

Water::~Water()
{
	Release(mMesh);
	Release(mWaterFX);
	Release(mWaveMap0);
	Release(mWaveMap1);
}

void Water::OnLostDevice()
{
	HR(mWaterFX->OnLostDevice());
}

void Water::OnResetDevice()
{
	HR(mWaterFX->OnResetDevice());
}

///////////////////////////////////////////////////////
//Use this function to generate the water grid mesh
///////////////////////////////////////////////////////
void Water::GenerateGrid(int numVertRows, int numVertCols, 
						 float dx, float dz, const D3DXVECTOR3 &center, 
						 std::vector<D3DXVECTOR3> &verts, std::vector<DWORD> &indices)
{
	int numVertices = numVertRows*numVertCols;
	int numCellRows = numVertRows-1;
	int numCellCols = numVertCols-1;

	int numTris = numCellRows*numCellCols*2;

	float width = (float)numCellCols * dx;
	float depth = (float)numCellRows * dz;

	//Build vertices
	verts.resize( numVertices );

	// Offsets to translate grid from quadrant 4 to center of 
	// coordinate system.
	float xOffset = -width * 0.5f; 
	float zOffset =  depth * 0.5f;

	int k = 0;
	for(float i = 0; i < numVertRows; ++i)
	{
		for(float j = 0; j < numVertCols; ++j)
		{
			// Negate the depth coordinate to put in quadrant four.  
			// Then offset to center about coordinate system.
			verts[k].x =  j * dx + xOffset;
			verts[k].z = -i * dz + zOffset;
			verts[k].y =  0.0f;

			// Translate so that the center of the grid is at the
			// specified 'center' parameter.
			D3DXMATRIX T;
			D3DXMatrixTranslation(&T, center.x, center.y, center.z);
			D3DXVec3TransformCoord(&verts[k], &verts[k], &T);
			
			++k; // Next vertex
		}
	}

	// Build indices
	indices.resize(numTris * 3);
	 
	// Generate indices for each quad.
	k = 0;
	for(DWORD i = 0; i < (DWORD)numCellRows; ++i)
	{
		for(DWORD j = 0; j < (DWORD)numCellCols; ++j)
		{
			indices[k]     =   i   * numVertCols + j;
			indices[k + 1] =   i   * numVertCols + j + 1;
			indices[k + 2] = (i+1) * numVertCols + j;
					
			indices[k + 3] = (i+1) * numVertCols + j;
			indices[k + 4] =   i   * numVertCols + j + 1;
			indices[k + 5] = (i+1) * numVertCols + j + 1;

			// next quad
			k += 6;
		}
	}
}

void Water::update(float dt)
{
	// Update texture coordinate offsets.  These offsets are added to the
	// texture coordinates in the vertex shader to animate them.
	mWaveMapOffset0 += mInitInfo.waveMapVelocity0 * dt;
	mWaveMapOffset1 += mInitInfo.waveMapVelocity1 * dt;

	// Textures repeat every 1.0 unit, so reset back down to zero
	// so the coordinates do not grow too large.
	if(mWaveMapOffset0.x >= 1.0f || mWaveMapOffset0.x <= -1.0f)
		mWaveMapOffset0.x = 0.0f;
	if(mWaveMapOffset1.x >= 1.0f || mWaveMapOffset1.x <= -1.0f)
		mWaveMapOffset1.x = 0.0f;
	if(mWaveMapOffset0.y >= 1.0f || mWaveMapOffset0.y <= -1.0f)
		mWaveMapOffset0.y = 0.0f;
	if(mWaveMapOffset1.y >= 1.0f || mWaveMapOffset1.y <= -1.0f)
		mWaveMapOffset1.y = 0.0f;
}

void Water::draw(D3DXMATRIX pMatrix, D3DXVECTOR3 CamPos)
{
	HR(mWaterFX->SetMatrix(mhWVP, &(mInitInfo.toWorld*pMatrix)));
	HR(mWaterFX->SetValue(mhEyePosW, &CamPos, sizeof(D3DXVECTOR3)));
	HR(mWaterFX->SetValue(mhWaveMapOffset0, &mWaveMapOffset0, sizeof(D3DXVECTOR2)));
	HR(mWaterFX->SetValue(mhWaveMapOffset1, &mWaveMapOffset1, sizeof(D3DXVECTOR2)));

	UINT numPasses = 0;
	HR(mWaterFX->Begin(&numPasses, 0));
	HR(mWaterFX->BeginPass(0));

		HR(mMesh->DrawSubset(0));

	HR(mWaterFX->EndPass());
	HR(mWaterFX->End());
}

void Water::buildFX()
{
	ID3DXBuffer* errors = 0;
	HR(D3DXCreateEffectFromFile(gd3dDevice, "shaders/water.c",
		0, 0, D3DXSHADER_DEBUG, 0, &mWaterFX, &errors));

	if(errors)
	{
		MessageBox(0, (char*)errors->GetBufferPointer(), 0, 0);
	}

 	mhTech           = mWaterFX->GetTechniqueByName("WaterTech");
	mhWorld          = mWaterFX->GetParameterByName(0, "gWorld");
	mhWorldInv       = mWaterFX->GetParameterByName(0, "gWorldInv");
	mhWVP            = mWaterFX->GetParameterByName(0, "gWVP");
	mhEyePosW        = mWaterFX->GetParameterByName(0, "gEyePosW");
	mhLight          = mWaterFX->GetParameterByName(0, "gLight");
	mhMtrl           = mWaterFX->GetParameterByName(0, "gMtrl");
	mhWaveMap0       = mWaterFX->GetParameterByName(0, "gWaveMap0");
    mhWaveMap1       = mWaterFX->GetParameterByName(0, "gWaveMap1");
	mhWaveMapOffset0 = mWaterFX->GetParameterByName(0, "gWaveMapOffset0");
	mhWaveMapOffset1 = mWaterFX->GetParameterByName(0, "gWaveMapOffset1");
	mhEnvMap         = mWaterFX->GetParameterByName(0, "gEnvMap");


	// We don't need to set these every frame since they do not change.
	HR(mWaterFX->SetMatrix(mhWorld, &mInitInfo.toWorld));
	D3DXMATRIX worldInv;
	D3DXMatrixInverse(&worldInv, 0, &mInitInfo.toWorld);
	HR(mWaterFX->SetMatrix(mhWorldInv, &worldInv));
	HR(mWaterFX->SetTechnique(mhTech));
	HR(mWaterFX->SetTexture(mhWaveMap0, mWaveMap0));
	HR(mWaterFX->SetTexture(mhWaveMap1, mWaveMap1));
	HR(mWaterFX->SetValue(mhLight, &mInitInfo.dirLight, sizeof(DirLight)));
	HR(mWaterFX->SetValue(mhMtrl, &mInitInfo.mtrl, sizeof(Mtrl)));
}

void Water::setEnvMap(IDirect3DCubeTexture9* envMap)
{
	HR(mWaterFX->SetTexture(mhEnvMap, envMap));
}